import dash
import pandas as pd
import plotly.graph_objects as go

from copy import deepcopy
from typing import List, Dict, Union, Optional, Mapping, Any
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
from webviz_ert.controllers.controller_functions import response_options
from webviz_ert.plugins import WebvizErtPluginABC
from webviz_ert.models import (
    ResponsePlotModel,
    BoxPlotModel,
    Response,
    MultiHistogramPlotModel,
    load_ensemble,
)
from webviz_ert import assets


def _get_univariate_misfits_boxplots(
    misfits_df: Optional[pd.DataFrame], ensemble_name: str, color: str
) -> List[BoxPlotModel]:
    if misfits_df is None:
        return []
    x_axis = misfits_df.columns
    realization_names = [f"Realization {name}" for name in misfits_df.index.values]
    misfit_plots = list()
    for misfits in misfits_df:
        plot = BoxPlotModel(
            y_axis=misfits_df[misfits].abs().values,
            name=misfits,
            ensemble_name=ensemble_name,
            color=color,
            customdata=realization_names,
            hovertemplate="(%{x}, %{y}) - %{customdata}",
        )
        misfit_plots.append(plot)
    return misfit_plots


def _create_misfits_plot(
    response: Response, selected_realizations: List[int], color: str, ensemble_name: str
) -> ResponsePlotModel:
    realizations = _get_univariate_misfits_boxplots(
        response.univariate_misfits_df(selected_realizations),
        ensemble_name=ensemble_name,
        color=color,
    )
    ensemble_plot = ResponsePlotModel(
        realizations,
        [],
        dict(
            hovermode="closest",
            uirevision=True,
        ),
    )
    return ensemble_plot


def observation_response_controller(parent: WebvizErtPluginABC, app: dash.Dash) -> None:
    @app.callback(
        [
            Output(parent.uuid("response-selector"), "options"),
            Output(parent.uuid("response-selector"), "value"),
            Output(parent.uuid("response-selector-store"), "data"),
        ],
        [
            Input(parent.uuid("selected-ensemble-dropdown"), "value"),
            Input(parent.uuid("response-selector"), "value"),
        ],
        [
            State(parent.uuid("response-selector-store"), "data"),
        ],
    )
    def set_response_callback(
        selected_ensembles: List[str],
        selected_resp: Optional[str],
        selected_resp_store: Optional[str],
    ) -> List[Any]:
        ctx = dash.callback_context
        triggered_id = ctx.triggered[0]["prop_id"].split(".")[0]
        if not triggered_id:
            raise PreventUpdate

        if triggered_id == parent.uuid("response-selector"):
            selected_resp_store = selected_resp

        ensembles = [
            load_ensemble(parent, ensemble_id) for ensemble_id in selected_ensembles
        ]
        responses = response_options(response_filters=["obs"], ensembles=ensembles)
        options = [{"label": name, "value": name} for name in sorted(responses)]

        if options:
            selected_resp = selected_resp_store
        else:
            selected_resp = None
            selected_resp_store = None

        parent.save_state("response", selected_resp_store)

        return [options, selected_resp, selected_resp_store]

    @app.callback(
        Output(
            {"id": parent.uuid("response-graphic"), "type": parent.uuid("graph")},
            "figure",
        ),
        [
            Input(parent.uuid("response-selector"), "value"),
            Input(parent.uuid("yaxis-type"), "value"),
            Input(parent.uuid("misfits-type"), "value"),
        ],
        [State(parent.uuid("selected-ensemble-dropdown"), "value")],
    )
    def update_graph(
        response: Optional[str],
        yaxis_type: List[str],
        misfits_type: str,
        selected_ensembles: List[str],
    ) -> go.Figure:
        if not response or response == "" or not selected_ensembles:
            return go.Figure()

        if misfits_type == "Summary":
            data_dict = {}
            colors = {}
            for index, ensemble_id in enumerate(selected_ensembles):
                ensemble = load_ensemble(parent, ensemble_id)
                summary_df = ensemble.responses[response].summary_misfits_df(
                    selection=None
                )  # What about selections?
                if summary_df is not None:
                    data_dict[ensemble.name] = summary_df.transpose()
                colors[ensemble.name] = assets.get_color(index=index)
            if data_dict:
                plot = MultiHistogramPlotModel(
                    data_dict,
                    names={name: name for name in data_dict},
                    colors=colors,
                    hist=True,
                    kde=False,
                )
                return plot.repr

        def _generate_plot(ensemble_id: str, color: str) -> ResponsePlotModel:
            ensemble = load_ensemble(parent, ensemble_id)
            resp = ensemble.responses[str(response)]
            plot = _create_misfits_plot(resp, [], color, ensemble.name)
            return plot

        response_plots = [
            _generate_plot(ensemble_id, assets.get_color(index=index))
            for index, ensemble_id in enumerate(selected_ensembles)
        ]

        fig = go.Figure()
        for plt in response_plots:
            for trace in plt.repr.data:
                fig.add_trace(trace)
        fig.update_yaxes(type=yaxis_type)
        return fig



from webviz_ert.models import Response


def test_response_model(mock_data):
    resp_model = Response(
        name="SNAKE_OIL_GPR_DIFF@199",
        ensemble_id="1",
        project_id="",
        ensemble_size=1,
        active_realizations=[0],
        resp_schema={"id": "id", "name": "name"},
    )
    assert resp_model.name == "SNAKE_OIL_GPR_DIFF@199"
    assert len(resp_model.data.columns) == 1



from .ensemble_selector_controller import ensemble_list_selector_controller
from .multi_response_controller import multi_response_controller
from .observation_response_controller import observation_response_controller
from .multi_parameter_controller import multi_parameter_controller
from .controller_functions import response_options, parameter_options
from .plot_view_controller import plot_view_controller
from .parameter_comparison_controller import parameter_comparison_controller
from .parameter_selector_controller import parameter_selector_controller
from .response_correlation_controller import response_correlation_controller
from .element_dropdown_controller import element_dropdown_controller



import dash

from typing import List, Tuple, Dict, Optional, Any
from dash.dependencies import Input, Output, State

from webviz_ert.models import DataType
from webviz_ert.plugins import WebvizErtPluginABC
from webviz_ert.models import load_ensemble
from webviz_ert.controllers import parameter_options, response_options


def element_dropdown_controller(
    parent: WebvizErtPluginABC,
    app: dash.Dash,
    data_type: DataType,
) -> None:
    @app.callback(
        [
            Output(parent.uuid(f"element-dropdown-{data_type}"), "options"),
            Output(parent.uuid(f"element-dropdown-{data_type}"), "value"),
            Output(parent.uuid(f"element-dropdown-store-{data_type}"), "data"),
        ],
        [
            Input(parent.uuid("ensemble-selection-store"), "modified_timestamp"),
            Input(parent.uuid(f"element-dropdown-{data_type}"), "value"),
        ],
        [
            State(parent.uuid(f"element-dropdown-store-{data_type}"), "data"),
            State(parent.uuid("ensemble-selection-store"), "data"),
        ],
    )
    def set_callbacks(
        _: Any,
        selected: Optional[str],
        selected_store: Optional[str],
        ensemble_selection_store: Dict[str, List],
    ) -> Tuple[List[Dict], Optional[str], Optional[str]]:
        if not ensemble_selection_store or not ensemble_selection_store["selected"]:
            return [], None, None

        selected_ensembles = [
            selection["value"] for selection in ensemble_selection_store["selected"]
        ]

        ensembles = [
            load_ensemble(parent, ensemble_id) for ensemble_id in selected_ensembles
        ]
        if data_type == DataType.PARAMETER:
            options = parameter_options(ensembles, union_keys=False)
        elif data_type == DataType.RESPONSE:
            options = response_options(response_filters=[], ensembles=ensembles)
        else:
            raise ValueError(f"Undefined data type {data_type}")

        element_options = [{"label": name, "value": name} for name in sorted(options)]

        # Keep track of selected element for the session
        if not selected and selected_store is not None:
            selected = selected_store

        return element_options, selected, selected



import dash

from typing import List, Any, Tuple, Dict, Optional
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State

from webviz_ert.plugins import WebvizErtPluginABC
from webviz_ert.models import (
    load_ensemble,
)
from webviz_ert.controllers import parameter_options, response_options
from webviz_ert.models.data_model import DataType


def _filter_match(_filter: str, key: str) -> bool:
    return _filter.lower() in key.lower()


def parameter_selector_controller(
    parent: WebvizErtPluginABC,
    app: dash.Dash,
    data_type: DataType,
    union_keys: bool = True,
    extra_input: bool = False,
) -> None:
    parameter_selector_multi_id = parent.uuid(f"parameter-selector-multi-{data_type}")
    parameter_selector_filter_id = parent.uuid(f"parameter-selector-filter-{data_type}")
    parameter_deactivator_id = parent.uuid(f"parameter-deactivator-{data_type}")
    parameter_selection_store_id = parent.uuid(f"parameter-selection-store-{data_type}")
    options_inputs = [
        Input(parent.uuid("selected-ensemble-dropdown"), "value"),
        Input(parameter_selector_filter_id, "value"),
        Input(parameter_deactivator_id, "value"),
    ]
    if extra_input:
        options_inputs.extend(
            [Input(parent.uuid("response-observations-check"), "value")]
        )

    @app.callback(
        [
            Output(parameter_selector_multi_id, "options"),
            Output(parameter_selector_multi_id, "value"),
        ],
        options_inputs,
    )
    def update_parameters_options(
        selected_ensembles: List[str],
        filter_search: str,
        selected: Optional[List[str]],
        *args: List[str],
    ) -> Tuple[List[Dict], Optional[List[str]]]:
        if not selected_ensembles:
            # Reset selection list
            return [], []

        response_filter = []
        if extra_input:
            response_filter = args[0]
        selected_set = set() if not selected else set(selected)
        ensembles = [
            load_ensemble(parent, ensemble_id) for ensemble_id in selected_ensembles
        ]
        if data_type == DataType.PARAMETER:
            options = parameter_options(ensembles, union_keys=union_keys)
        elif data_type == DataType.RESPONSE:
            options = response_options(response_filter, ensembles)
        else:
            raise ValueError(f"Undefined parameter type {data_type}")

        options = options.difference(selected_set)
        if filter_search:
            options = {name for name in options if _filter_match(filter_search, name)}

        return [{"label": name, "value": name} for name in sorted(options)], selected

    @app.callback(
        Output(parameter_selection_store_id, "data"),
        [
            Input(parameter_selector_multi_id, "value"),
            Input(parameter_selector_filter_id, "n_submit"),
            Input(parent.uuid("selected-ensemble-dropdown"), "value"),
        ],
        State(parameter_deactivator_id, "value"),
    )
    def update_parameter_selection(
        parameters: List[str],
        _: int,
        selected_ensembles: List[str],
        selected_params: Optional[List[str]],
    ) -> Optional[List[str]]:
        stored_parameters = None
        ctx = dash.callback_context
        triggered_id = ctx.triggered[0]["prop_id"].split(".")[0]
        if triggered_id == parameter_selector_filter_id:
            # Prevent selecting everything from the search result on enter
            raise PreventUpdate
        elif triggered_id == parent.uuid("selected-ensemble-dropdown"):
            if selected_ensembles is None or selected_ensembles == []:
                stored_parameters = []
            else:
                raise PreventUpdate
        elif triggered_id == parameter_selector_multi_id:
            selected_params = [] if not selected_params else selected_params
            parameters = (
                []
                if not parameters
                else [
                    parameter
                    for parameter in parameters
                    if parameter not in selected_params
                ]
            )
            stored_parameters = selected_params + parameters

        parent.save_state(f"{data_type}", stored_parameters)

        return stored_parameters

    @app.callback(
        [
            Output(parameter_deactivator_id, "options"),
            Output(parameter_deactivator_id, "value"),
        ],
        [
            Input(parameter_selection_store_id, "modified_timestamp"),
        ],
        [
            State(parameter_selection_store_id, "data"),
        ],
    )
    def update_parameter_options(
        _: Any, shown_parameters: Optional[List[str]]
    ) -> Tuple[List[Dict], List[str]]:
        shown_parameters = [] if not shown_parameters else shown_parameters
        selected_opts = [{"label": param, "value": param} for param in shown_parameters]
        return selected_opts, shown_parameters

    container_parameter_selector_multi_id = parent.uuid(
        f"container-parameter-selector-multi-{data_type}"
    )
    parameter_selector_button_id = parent.uuid(f"parameter-selector-button")

    @app.callback(
        Output(container_parameter_selector_multi_id, "className"),
        [
            Input(parameter_selector_button_id, "n_clicks"),
        ],
        [
            State(container_parameter_selector_multi_id, "className"),
        ],
    )
    def toggle_selector_visibility(_: int, class_name: str) -> str:
        ctx = dash.callback_context
        triggered_id = ctx.triggered[0]["prop_id"].split(".")[0]
        if triggered_id == parameter_selector_button_id:
            if class_name == "ert-parameter-selector-container-hide":
                class_name = "ert-parameter-selector-container-show"
            else:
                class_name = "ert-parameter-selector-container-hide"
        return class_name



import argparse
import logging
import os
import pathlib
import shutil
import signal
import sys
import tempfile
from typing import Any, Dict, Optional

import yaml

from webviz_ert.assets import WEBVIZ_CONFIG

logger = logging.getLogger()


def run_webviz_ert(
    title: str,
    experimental_mode: bool = False,
    verbose: bool = False,
    project_identifier: Optional[str] = None,
) -> None:
    signal.signal(signal.SIGINT, handle_exit)
    # The entry point of webviz is to call it from command line, and so do we.

    webviz = shutil.which("webviz")
    if webviz:
        send_ready()
        with tempfile.NamedTemporaryFile() as temp_config:
            if project_identifier is None:
                project_identifier = os.getcwd()
            create_config(
                title, project_identifier, WEBVIZ_CONFIG, temp_config, experimental_mode
            )
            os.execl(
                webviz,
                webviz,
                "build",
                temp_config.name,
                "--theme",
                "equinor",
                "--loglevel",
                "DEBUG" if verbose else "WARNING",
            )
    else:
        logger.error("Failed to find webviz")


def send_ready() -> None:
    """
    Tell ERT's BaseService that we're ready, even though we're not actually
    ready to accept requests. At the moment, ERT doesn't interface with
    webviz-ert in any way, so it's not necessary to send the signal later.
    """
    if "ERT_COMM_FD" not in os.environ:
        logger.info("webviz ert is running outside of ert context")
        return
    fd = int(os.environ["ERT_COMM_FD"])
    with os.fdopen(fd, "w") as f:
        f.write("{}")  # Empty, but valid JSON


def handle_exit(
    *_: Any,
) -> None:
    # pylint: disable=logging-not-lazy
    logger.info("\n" + "=" * 32)
    logger.info("Session terminated by the user.\nThank you for using webviz-ert!")
    logger.info("=" * 32)
    sys.tracebacklimit = 0
    sys.stdout = open(os.devnull, "w")
    sys.exit()


def create_config(
    title: str,
    project_identifier: Optional[str],
    config_file: pathlib.Path,
    temp_config: Any,
    experimental_mode: bool,
) -> None:
    with open(config_file, "r") as f:
        config_dict = yaml.safe_load(f)
        for page in config_dict["pages"]:
            for element in page["content"]:
                for key in element:
                    if element[key] is None:
                        element[key] = {"project_identifier": project_identifier}
                    else:
                        element[key]["project_identifier"] = project_identifier

    new_config_dict = config_dict

    def filter_experimental_pages(
        page: Dict[str, Any], experimental_mode: bool
    ) -> bool:
        if "experimental" in page:
            if page["experimental"]:
                return experimental_mode
        return True

    new_config_dict["pages"] = [
        page
        for page in new_config_dict["pages"]
        if filter_experimental_pages(page, experimental_mode)
    ]

    new_config_dict["title"] = f"{title} - {project_identifier}"

    output_str = yaml.dump(new_config_dict)
    temp_config.write(str.encode(output_str))
    temp_config.seek(0)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--experimental-mode", action="store_true")
    parser.add_argument(
        "--verbose", action="store_true", help="Show verbose output.", default=False
    )
    parser.add_argument(
        "--title", help="Set title of html document", default="ERT - Visualization tool"
    )
    parser.add_argument(
        "--project_identifier", help="Set title of html document", default=os.getcwd()
    )

    args = parser.parse_args()

    run_webviz_ert(
        args.title, args.experimental_mode, args.verbose, args.project_identifier
    )



import pandas as pd
import pytest

from webviz_ert.controllers.response_correlation_controller import (
    _define_style_ensemble,
    _layout_figure,
    _format_index_value,
    _format_index_text,
)

from webviz_ert.controllers.response_correlation_controller import (
    _sort_dataframe,
    _get_selected_indexes,
)
from webviz_ert.models import PlotModel


PLOT_STYLE = {
    "mode": "markers",
    "line": None,
    "marker": {"color": "rgba(56,108,176,0.8)", "size": 9},
    "text": "NA",
}


def test_sort_dataframe():
    one_key = "WOPR"
    other_key = "BGMC"
    data = {}
    data[one_key] = [-0.4, 0.6, 0.2]
    data[other_key] = [0.3, 0.8, -0.1]
    dataframe = pd.DataFrame(data=data)
    index = None
    sorted_dataframe, _ = _sort_dataframe(dataframe, index, other_key)
    assert list(sorted_dataframe[other_key]) == sorted(data[other_key])
    assert list(sorted_dataframe[one_key]) == [0.2, -0.4, 0.6]


@pytest.mark.parametrize(
    "index,expected_color", [(0, "rgba(56,108,176,0.8)"), (1, "rgba(127,201,127,0.8)")]
)
def test_define_style_ensemble_color(index, expected_color):
    x_axis = pd.Index([1])
    style = _define_style_ensemble(index, x_axis)
    assert style["line"]["color"] == expected_color
    assert style["marker"]["color"] == expected_color


@pytest.mark.parametrize(
    "x_axis_list,expected_mode",
    [
        (
            [pd.Timestamp("01-01-2020")],
            "markers+lines",
        ),
        ([str(1)], "markers"),
    ],
)
def test_define_style_ensemble_mode(x_axis_list, expected_mode):
    x_axis = pd.Index(x_axis_list)
    style = _define_style_ensemble(0, x_axis)
    assert style["mode"] == expected_mode


def test_layout_figure():
    layout = _layout_figure(x_axis_label="Index")  # "Date" is also possible
    expected_layout = {
        "hovermode": "closest",
        "uirevision": True,
        "margin": {"l": 5, "b": 5, "t": 25, "r": 5},
        "autosize": True,
        "showlegend": False,
        "clickmode": "event+select",
        "xaxis": {"title": {"text": "Index"}},
        "yaxis": {"title": {"text": "Value"}},
    }

    assert layout == expected_layout


@pytest.mark.parametrize(
    "raw_value,expected_formatted_value",
    [
        ("2022-08-05 14:25:00", "2022-08-05"),
        ("2022-09-21 14:25:00", "2022-09-21"),
        (14, "14"),
        ("213", "213"),
        ("SPAM", "SPAM"),
    ],
    ids=["date1", "date2", "numeric", "numeric-as-string", "silly-string"],
)
def test_format_index_text(raw_value: str, expected_formatted_value: str):
    assert _format_index_text(raw_value) == expected_formatted_value


@pytest.mark.parametrize(
    "raw_value,expected_formatted_value",
    [
        ("2022-08-05 14:25:00", "2022-08-05 14:25:00"),
        ("2022-09-21 14:25:00", "2022-09-21 14:25:00"),
        (14, 14),
        ("213", 213),
        ("SPAM", "SPAM"),
    ],
    ids=["date1", "date2", "numeric", "numeric-as-string", "silly-string"],
)
def test_format_index_value(raw_value: str, expected_formatted_value: str):
    assert _format_index_value(raw_value) == expected_formatted_value


@pytest.mark.parametrize(
    "plots, date_ranges, expected",
    [
        ([], None, {}),
        ([], {}, {}),
        ([], {}, {}),
        (
            [
                PlotModel(
                    x_axis=[0, 5, 10], y_axis=[1, 1, 1], name="plot1", **PLOT_STYLE
                )
            ],
            {"x": ["2020-01-01", "2020-01-03"]},
            {},
        ),
        (
            [
                PlotModel(
                    x_axis=[0, 5, 10], y_axis=[1, 1, 1], name="plot1", **PLOT_STYLE
                )
            ],
            {"x2": [1, 3]},
            {},
        ),
        (
            [
                PlotModel(
                    x_axis=[0, 5, 10], y_axis=[1, 1, 1], name="plot1", **PLOT_STYLE
                ),
                PlotModel(
                    x_axis=[pd.Timestamp("2020-01-02")],
                    y_axis=[1, 1, 1],
                    name="plot2",
                    **PLOT_STYLE
                ),
            ],
            {"x": ["2020-01-01", "2020-01-03"]},
            {"plot2": [pd.Timestamp("2020-01-02")]},
        ),
        (
            [
                PlotModel(
                    x_axis=[0, 5, 10], y_axis=[1, 1, 1], name="plot1", **PLOT_STYLE
                ),
                PlotModel(
                    x_axis=[pd.Timestamp("2020-01-02")],
                    y_axis=[1, 1, 1],
                    name="plot2",
                    **PLOT_STYLE
                ),
            ],
            {"x": ["2020-01-01", "2020-01-03"], "x2": [4, 6]},
            {"plot1": [5], "plot2": [pd.Timestamp("2020-01-02")]},
        ),
    ],
)
def test_get_selected_indexes(plots, date_ranges, expected):
    assert _get_selected_indexes(plots, date_ranges) == expected



from typing import List, Set
from webviz_ert.models import Response, EnsembleModel


def _valid_response_option(response_filters: List[str], response: Response) -> bool:
    if "historical" in response_filters:
        if response.name.split(":")[0][-1] == "H":
            return False

    if "obs" in response_filters:
        return response.has_observations

    return True


def response_options(
    response_filters: List[str],
    ensembles: List[EnsembleModel],
) -> Set:
    response_names = set()

    for ensemble in ensembles:
        for name, response in ensemble.responses.items():
            if name in response_names:
                continue
            if _valid_response_option(response_filters, response):
                response_names.add(name)
    return response_names


def parameter_options(ensembles: List[EnsembleModel], union_keys: bool = True) -> Set:
    params_included: Set[str] = set()
    for ensemble in ensembles:
        if ensemble.parameters:
            parameters = set([parameter for parameter in ensemble.parameters])
            if not params_included:
                params_included = parameters
            elif union_keys:
                params_included = params_included.union(parameters)
            else:
                params_included = params_included.intersection(parameters)
    return params_included



from dash.development.base_component import Component
from webviz_config import WebvizPluginABC

from dash import html
from dash import dcc
from webviz_ert.models.data_model import DataType


def element_dropdown_view(parent: WebvizPluginABC, data_type: DataType) -> Component:
    return html.Div(
        [
            dcc.Dropdown(
                id=parent.uuid(f"element-dropdown-{data_type}"),
                multi=False,
            ),
            dcc.Store(
                id=parent.uuid(f"element-dropdown-store-{data_type}"),
                storage_type="session",
                data=None,
            ),
        ],
    )



import warnings

with warnings.catch_warnings():
    warnings.filterwarnings(
        "ignore",
        message=r".*pkg_resources is deprecated as an API.*",
        category=UserWarning,
    )
    import pkg_resources
import json
from pathlib import Path
from webviz_config.webviz_assets import WEBVIZ_ASSETS


ASSETS_DIR = Path(pkg_resources.resource_filename("webviz_ert", "assets"))
WEBVIZ_ASSETS.add(ASSETS_DIR / "bootstrap-grid.css")
WEBVIZ_ASSETS.add(ASSETS_DIR / "bootstrap.min.css")
WEBVIZ_ASSETS.add(ASSETS_DIR / "ert-style.css")
with open(ASSETS_DIR / "ert-style.json") as f:
    ERTSTYLE = json.load(f)

COLOR_WHEEL = ERTSTYLE["ensemble-selector"]["color_wheel"]

WEBVIZ_CONFIG = (
    Path(pkg_resources.resource_filename("webviz_ert", "assets")) / "webviz-config.yml"
)


def get_color(index: int) -> str:
    return COLOR_WHEEL[index % len(COLOR_WHEEL)]



from typing import MutableMapping, Optional, Dict, Any
import dash
from webviz_config import WebvizPluginABC
from webviz_ert.models import EnsembleModel
import pathlib
import json
import tempfile


class WebvizErtPluginABC(WebvizPluginABC):
    _ensembles: MutableMapping[str, "EnsembleModel"] = {}
    _state: MutableMapping[str, Any] = {}
    _state_file_name: str = "webviz_ert_state.json"
    _state_path: Optional[pathlib.Path] = None

    def __init__(self, app: dash.Dash, project_identifier: str):
        super().__init__()
        if not project_identifier:
            project_identifier = tempfile.NamedTemporaryFile().name
            WebvizErtPluginABC._state = {}
            WebvizErtPluginABC._state_path = None

        self.project_identifier: str = project_identifier
        WebvizErtPluginABC._state = self.init_state(pathlib.Path(project_identifier))
        self._class_name: str = type(self).__name__.lower()

    @staticmethod
    def init_state(project_root: pathlib.Path) -> MutableMapping[str, Any]:
        if not WebvizErtPluginABC._state:
            if WebvizErtPluginABC._state_path is None:
                WebvizErtPluginABC._state_path = (
                    project_root / f"{WebvizErtPluginABC._state_file_name}"
                )
            if not WebvizErtPluginABC._state_path.exists():
                WebvizErtPluginABC._state_path.parent.mkdir(parents=True, exist_ok=True)
                return dict()
            with open(WebvizErtPluginABC._state_path, "r", encoding="utf-8") as f:
                return json.load(f)
        return WebvizErtPluginABC._state

    @classmethod
    def get_ensembles(cls) -> MutableMapping[str, "EnsembleModel"]:
        return cls._ensembles

    @classmethod
    def get_ensemble(cls, ensemble_id: str) -> Optional[EnsembleModel]:
        return cls._ensembles.get(ensemble_id)

    @classmethod
    def add_ensemble(cls, ensemble: "EnsembleModel") -> None:
        if ensemble.id not in cls._ensembles:
            cls._ensembles[ensemble.id] = ensemble

    @classmethod
    def clear_ensembles(cls) -> None:
        cls._ensembles.clear()

    def save_state(self, key: str, data: Any) -> None:
        page_state = WebvizErtPluginABC._state.get(self._class_name, {})
        page_state[key] = data
        WebvizErtPluginABC._state[self._class_name] = page_state
        if WebvizErtPluginABC._state_path is not None:
            with open(WebvizErtPluginABC._state_path, "w", encoding="utf-8") as f:
                json.dump(WebvizErtPluginABC._state, f, indent=4, sort_keys=True)

    def load_state(
        self, key: Optional[str] = None, default: Optional[Any] = None
    ) -> MutableMapping[str, Any]:
        if key is not None:
            page_state = WebvizErtPluginABC._state.get(self._class_name, {})
            return page_state.get(key, default)

        return WebvizErtPluginABC._state



import pandas as pd
from typing import List, Any, Optional, Union
from webviz_ert.data_loader import get_data_loader


class PriorModel:
    def __init__(
        self,
        function: str,
        function_parameter_names: List[str],
        function_parameter_values: List[Union[float, int]],
    ):
        self.function = function
        self.function_parameter_names = function_parameter_names
        self.function_parameter_values = function_parameter_values


class ParametersModel:
    def __init__(self, **kwargs: Any):
        self._project_id = kwargs["project_id"]
        self.group = kwargs["group"]
        self.key = kwargs["key"]
        self.priors = kwargs["prior"]
        self._id = kwargs["param_id"]
        self._ensemble_id = kwargs["ensemble_id"]
        self._realizations = kwargs.get("realizations")
        self._data_df = pd.DataFrame()
        self._data_loader = get_data_loader(self._project_id)

    def data_df(self) -> pd.DataFrame:
        if self._data_df.empty:
            _data_df = self._data_loader.get_ensemble_parameter_data(
                ensemble_id=self._ensemble_id,
                parameter_name=self.key,
            )
            if _data_df is not None:
                _data_df.index.name = self.key
                self._data_df = _data_df
        return self._data_df



from typing import List
from dash.development.base_component import Component
from webviz_ert.plugins import WebvizErtPluginABC

from dash import html
from dash import dcc
import webviz_core_components as wcc
import dash_bootstrap_components as dbc


def ensemble_selector_list(parent: WebvizErtPluginABC) -> List[Component]:
    return [
        html.Div(
            dbc.Row(
                [
                    dbc.Col(
                        html.H6("Ensembles", className="ert-label"),
                        align="left",
                        width="auto",
                    ),
                    dbc.Col(
                        html.Button(
                            id=parent.uuid("ensemble-refresh-button"),
                            children="Refresh",
                            n_clicks=0,
                        ),
                        align="right",
                    ),
                    dbc.Col(
                        html.Button(
                            id=parent.uuid("parameter-selector-button"),
                            children=("Hide Selectors"),
                        ),
                        align="right",
                        width="auto",
                    ),
                ],
                align="center",
            )
        ),
        html.Div(
            wcc.Select(
                id=parent.uuid("ensemble-multi-selector"),
                multi=True,
                size=10,
                persistence=True,
                persistence_type="session",
                options=[],
                value=[],
                className="ert-select-variable",
            ),
            id=parent.uuid("container-ensemble-selector-multi"),
            className="ert-ensemble-selector-container-show",
        ),
        dcc.Dropdown(
            id=parent.uuid("selected-ensemble-dropdown"),
            value=[],
            multi=True,
            options=[],
            persistence=True,
            searchable=False,
            placeholder="",
            persistence_type="session",
            className="selected-ensemble-dropdown",
        ),
        dcc.Store(
            id=parent.uuid("ensemble-selection-store"),
            storage_type="session",
        ),
    ]



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setup(
    name="webviz-ert",
    description="Webviz plugins for ERT",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Equinor",
    packages=find_packages(exclude=["tests"]),
    package_data={"webviz_ert.assets": ["*"]},
    include_package_data=True,
    entry_points={
        "webviz_config_plugins": [
            "ResponseComparison = webviz_ert.plugins:ResponseComparison",
            "ObservationAnalyzer = webviz_ert.plugins:ObservationAnalyzer",
            "ParameterComparison = webviz_ert.plugins:ParameterComparison",
            "ResponseCorrelation = webviz_ert.plugins:ResponseCorrelation",
        ],
    },
    install_requires=[
        "dash-bootstrap-components",
        "dash-daq",
        "dash<3.0.0",
        "pandas",
        "requests",
        "webviz-config-equinor",
        "webviz-config>=0.0.40",
        "webviz-subsurface-components",
    ],
    setup_requires=["setuptools_scm"],
    use_scm_version=True,
    zip_safe=False,
    classifiers=[
        "Natural Language :: English",
        "Environment :: Web Environment",
        "Framework :: Dash",
        "Framework :: Flask",
        "Topic :: Scientific/Engineering",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    ],
)



import pandas as pd
import plotly.graph_objects as go
import dash
import numpy as np
import dash_bootstrap_components as dbc

from typing import List, Optional, Dict, Tuple, Any, Union
from copy import deepcopy
from dash import html
from dash.development.base_component import Component
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State
from plotly.subplots import make_subplots
from plotly.graph_objs.layout import XAxis, YAxis

from webviz_ert.plugins import WebvizErtPluginABC
from webviz_ert.models import load_ensemble, BarChartPlotModel, PlotModel, AxisType
from webviz_ert.controllers.multi_response_controller import (
    _get_observation_plots,
    axis_label_for_ensemble_response,
)
from webviz_ert import assets


def response_correlation_controller(parent: WebvizErtPluginABC, app: dash.Dash) -> None:
    @app.callback(
        [
            Output(
                parent.uuid("response-correlation"),
                "figure",
            ),
            Output(
                parent.uuid("response-heatmap"),
                "figure",
            ),
        ],
        [
            Input(parent.uuid("ensemble-selection-store"), "data"),
            Input(parent.uuid("correlation-store-selected-obs"), "data"),
            Input(parent.uuid("correlation-store-active-resp-param"), "data"),
            Input(parent.uuid("correlation-metric"), "value"),
            Input(parent.uuid("sort-parameters"), "on"),
            Input(parent.uuid("hide-hover"), "on"),
        ],
        [
            State(parent.uuid("parameter-selection-store-param"), "data"),
            State(parent.uuid("parameter-selection-store-resp"), "data"),
        ],
    )
    def update_correlation_plot(
        ensemble_selection_store: Dict[str, List],
        selected_obs: Dict,
        active_resp_param: Dict,
        correlation_metric: str,
        sort_parameters: bool,
        hide_hover: bool,
        parameters: List[str],
        responses: List[str],
    ) -> Optional[Tuple[go.Figure, go.Figure]]:
        ensembles = _get_selected_ensembles_from_store(ensemble_selection_store)
        if not (
            ensembles
            and selected_obs
            and parameters
            and responses
            and active_resp_param["response"]["name"] in responses
        ):
            return {}, {}

        active_response = active_resp_param["response"]
        active_response_name = active_response["name"]
        active_index = active_response.get("index")
        if active_index is not None:
            active_response_name = f"{active_response_name}::{active_index}"
        data = {}
        colors = {}
        heatmaps = []

        df_index = None
        for index, ensemble_id in enumerate(ensembles):
            ensemble = load_ensemble(parent, ensemble_id)
            parameter_df = ensemble.parameters_df(parameters)
            if parameter_df.empty:
                continue
            valid_responses = []
            for response_name in responses:
                response = ensemble.responses[response_name]
                if response.empty:
                    continue
                resp_indexes = []
                valid_response_with_index = []
                if response.observations:
                    response_axis = response.axis
                    for obs_idx in selected_obs.get(response.name, []):
                        if obs_idx in response_axis:
                            resp_indexes.append(obs_idx)
                            valid_response_with_index.append(
                                f"{response.name}::{obs_idx}"
                            )

                    valid_responses += valid_response_with_index
                    new_dataframe = response.data.loc[resp_indexes].T
                    new_dataframe.columns = valid_response_with_index
                    parameter_df = pd.concat([parameter_df, new_dataframe], axis=1)

            corrdf = parameter_df.corr(method=correlation_metric)
            corrdf = corrdf.drop(valid_responses, axis=0).fillna(0)
            if active_response_name not in valid_responses:
                continue
            if sort_parameters:
                corrdf, df_index = _sort_dataframe(
                    corrdf, df_index, active_response_name
                )
            # create heatmap
            data_heatmap = {
                "type": "heatmap",
                "x": corrdf[valid_responses].columns,
                "y": corrdf[valid_responses].index,
                "z": [
                    corrdf[valid_responses].loc[parameter].values
                    for parameter in corrdf.index
                ],
                "zmin": -1,
                "zmax": 1,
                "colorscale": "rdylbu",
                "reversescale": True,
            }
            heatmaps.append(data_heatmap)
            ens_key = repr(ensemble)
            data[ens_key] = corrdf[active_response_name]
            data[ens_key].index.name = active_response_name
            colors[ens_key] = assets.get_color(index=index)
        if data and heatmaps:
            correlation_plot = BarChartPlotModel(data, colors)
            heatmap_plot = make_subplots(
                rows=1,
                cols=len(heatmaps),
                subplot_titles=[f"ENS_{idx + 1}" for idx, _ in enumerate(heatmaps)],
            )

            for idx, heatmap in enumerate(heatmaps):
                heatmap_plot.add_trace(heatmap, 1, 1 + idx)
                heatmap_plot.update_yaxes(showticklabels=False, row=1, col=1 + idx)
                heatmap_plot.update_xaxes(
                    showticklabels=True,
                    tickangle=90,
                    ticklabelposition="outside left",
                    row=1,
                    col=1 + idx,
                )
                if hide_hover:
                    heatmap_plot.update_traces(patch={"hoverinfo": "none"})
                else:
                    heatmap_plot.update_traces(
                        patch={"hovertemplate": "%{x} - %{y}: %{z}<extra></extra>"}
                    )
            _layout = assets.ERTSTYLE["figure"]["layout"].copy()

            _layout.update(
                {
                    "clickmode": "event+select",
                    "showlegend": False,
                    "annotations": [{"font": {"color": colors[ens]}} for ens in colors],
                    "hoverlabel": {
                        "font": {
                            "size": 10,
                        },
                    },
                }
            )
            heatmap_plot.update_layout(_layout)
            return correlation_plot.repr, heatmap_plot
        raise PreventUpdate

    @app.callback(
        [
            Output(parent.uuid("obs_index_selector_container"), "style"),
            Output(parent.uuid("obs_index_selector"), "figure"),
            Output(parent.uuid("correlation-store-selected-obs"), "data"),
            Output(parent.uuid("correlation-store-obs-range"), "data"),
        ],
        [
            Input(parent.uuid("parameter-selection-store-resp"), "data"),
            Input(parent.uuid("ensemble-selection-store"), "data"),
            Input(parent.uuid("obs_index_selector"), "selectedData"),
        ],
        [
            State(parent.uuid("correlation-store-obs-range"), "data"),
        ],
    )
    def update_observation_index_selector(
        responses: Optional[List[str]],
        ensemble_selection_store: Dict[str, List],
        selected_data: Optional[Dict],
        store_obs_range: Dict,
    ) -> Optional[Tuple[Dict, go.Figure, Dict, Dict]]:
        ctx = dash.callback_context
        triggered_id = ctx.triggered[0]["prop_id"].split(".")[0]

        def disable_figure() -> Tuple[Dict, go.Figure, Dict, Dict]:
            parent.save_state("correlation-store-obs-range", {})
            return {"display": "none", "min-height": "50px"}, go.Figure(), {}, {}

        ensembles = _get_selected_ensembles_from_store(ensemble_selection_store)
        if not (ensembles and responses):
            return disable_figure()
        if triggered_id == parent.uuid("obs_index_selector") and selected_data is None:
            store_obs_range = {}

        if selected_data is None or "range" not in selected_data:
            selected_data = {"range": store_obs_range}

        # Use only one ensemble
        ensemble = load_ensemble(parent, ensembles[0])
        obs_plots: List[PlotModel] = []
        index_axis = False
        time_axis = False
        for idx, response_name in enumerate(responses):
            response = ensemble.responses[response_name]
            if response.observations:
                for obs in response.observations:
                    if not obs.axis:
                        continue
                    if not index_axis and obs.axis_type == AxisType.INDEX:
                        index_axis = True
                    elif not time_axis and obs.axis_type == AxisType.TIMESTAMP:
                        time_axis = True
                    obs_plots.append(_observation_index_plot(obs, response_name, idx))

        if not obs_plots:
            return disable_figure()

        fig = go.Figure(layout_yaxis_range=[-1, len(obs_plots)])
        layout = assets.ERTSTYLE["figure"]["layout"].copy()
        layout.update(
            dict(
                xaxis=XAxis(title="Date", showgrid=False, fixedrange=True),
                xaxis2=XAxis(
                    overlaying="x",
                    title="Index",
                    side="top",
                    showgrid=False,
                    fixedrange=True,
                ),
                yaxis=YAxis(fixedrange=True),
                dragmode="select",
                template="plotly_white",
                height=300,
            )
        )
        fig.update_layout(layout)

        for plot in obs_plots:
            fig.add_trace(plot.repr)

        store_obs_range = selected_data["range"]
        selected_indexes = _get_selected_indexes(obs_plots, store_obs_range)

        # Add selection rectangle to figure
        x_range = store_obs_range.get("x")
        if not time_axis:
            x_range = store_obs_range.get("x2")
        if index_axis and "x2" not in store_obs_range:
            x_range = None

        if selected_indexes and x_range:
            x_start, x_end = x_range
            fig.add_shape(
                type="rect",
                yref="paper",
                x0=x_start,
                y0=0,
                x1=x_end,
                y1=1,
                fillcolor="PaleTurquoise",
                opacity=0.35,
            )
        else:
            # Use first index if nothing is selected
            for p in obs_plots:
                selected_indexes[p.name] = [p.x_axis[0]]

        parent.save_state("correlation-store-obs-range", store_obs_range)
        return (
            {"min-height": "300px", "padding-top": "20px"},
            fig,
            selected_indexes,
            store_obs_range,
        )

    @app.callback(
        Output(
            parent.uuid("response-overview"),
            "figure",
        ),
        [
            Input(parent.uuid("ensemble-selection-store"), "modified_timestamp"),
            Input(
                parent.uuid("correlation-store-active-resp-param"), "modified_timestamp"
            ),
        ],
        [
            State(parent.uuid("ensemble-selection-store"), "data"),
            State(parent.uuid("correlation-store-active-resp-param"), "data"),
        ],
    )
    def update_response_overview_plot(
        _: Any,
        __: Any,
        ensemble_selection_store: Dict[str, List],
        active_resp_param: Dict,
    ) -> Optional[go.Figure]:
        ensembles = _get_selected_ensembles_from_store(ensemble_selection_store)

        if not (ensembles and active_resp_param["response"]):
            return {}
        active_response = active_resp_param["response"]
        loaded_ensembles = [
            load_ensemble(parent, ensemble_id) for ensemble_id in ensembles
        ]

        response_plots = []
        obs_plots: List[PlotModel] = []

        for index, ensemble in enumerate(loaded_ensembles):
            response = ensemble.responses[active_response["name"]]
            response_x_axis = response.axis

            if isinstance(response_x_axis, pd.Index) and response_x_axis.empty:
                continue

            if response_x_axis is not None:
                style = _define_style_ensemble(index, response_x_axis)

            response_df = response.data_df().copy()
            response_plots += [
                PlotModel(
                    x_axis=response_x_axis,
                    y_axis=response_df[realization],
                    text=realization,
                    name=realization,
                    **style,
                )
                for realization in response_df
            ]

            if response.observations:
                for obs in response.observations:
                    observation_dataframe = obs.data_df()
                    metadata = len(observation_dataframe) * ["observation"]
                    obs_plots.append(
                        _get_observation_plots(observation_dataframe, metadata)
                    )

        fig = go.Figure()
        for plot in response_plots:
            fig.add_trace(plot.repr)

        x_axis_label = axis_label_for_ensemble_response(
            loaded_ensembles[0], active_response["name"]
        )

        fig.update_layout(_layout_figure(x_axis_label))

        selected_obs_idx = active_response.get("index")
        if selected_obs_idx is not None:
            fig.add_vline(
                x=selected_obs_idx,
                line=dict(color="red", dash="dash", width=4),
            )

        for plot in obs_plots:
            fig.add_trace(plot.repr)

        return fig

    @app.callback(
        Output(
            parent.uuid("response-scatterplot"),
            "figure",
        ),
        [
            Input(parent.uuid("correlation-store-active-resp-param"), "data"),
            Input(parent.uuid("ensemble-selection-store"), "data"),
            Input(parent.uuid("correlation-store-selected-obs"), "data"),
        ],
        [
            State(parent.uuid("parameter-selection-store-param"), "data"),
            State(parent.uuid("parameter-selection-store-resp"), "data"),
        ],
    )
    def update_response_parameter_scatterplot(
        active_resp_param: Dict,
        ensemble_selection_store: Dict[str, List],
        _: Dict,
        parameters: List[str],
        responses: List[str],
    ) -> go.Figure:
        ensembles = _get_selected_ensembles_from_store(ensemble_selection_store)
        if not (
            parameters
            and ensembles
            and responses
            and active_resp_param["parameter"] in parameters
            and active_resp_param["response"]
            and active_resp_param["response"]["name"] in responses
        ):
            return {}

        active_parameter = active_resp_param["parameter"]
        active_response = active_resp_param["response"]
        _plots = []
        _resp_plots = {}
        _param_plots = {}
        _colors = {}
        for index, ensemble_id in enumerate(ensembles):
            ensemble = load_ensemble(parent, ensemble_id)
            if ensemble.parameters and ensemble.responses:
                y_data = ensemble.parameters[active_parameter].data_df()
                response = ensemble.responses[active_response["name"]]
                if response.empty:
                    continue
                style = deepcopy(assets.ERTSTYLE["response-plot"]["response-index"])
                ensemble_color = assets.get_color(index=index)
                style["marker"]["color"] = ensemble_color
                _colors[str(ensemble)] = ensemble_color
                active_index: Optional[Any] = active_response.get("index")
                if response.observations:
                    response_axis = response.axis
                    if active_index is not None and active_index in response_axis:
                        x_data = response.data.loc[active_index]
                        _plots += [
                            PlotModel(
                                x_axis=x_data.values.flatten(),
                                y_axis=y_data.values.flatten(),
                                name=f"{ensemble.name}",
                                hoverlabel={"namelength": -1},
                                **style,
                            )
                        ]
                        _resp_plots[str(ensemble)] = x_data.values.flatten()
                        _param_plots[str(ensemble)] = y_data.values.flatten()

        n_rows = 10
        histogram_row_span = 2
        spacer_row_span = 1
        row_histograms = n_rows - (histogram_row_span // 2)

        specs = _create_scatterplot_specs(n_rows, histogram_row_span, spacer_row_span)
        fig = make_subplots(
            rows=n_rows,
            cols=2,
            specs=specs,
        )
        active_index_text = "NA"
        active_index = active_response.get("index")
        if active_index is not None:
            active_index_text = _format_index_text(active_index)
        _set_scatterplot_axes_titles(
            fig, row_histograms, active_response, active_parameter, active_index_text
        )

        _style_scatterplot(fig)

        _add_plots_to_multiplot(
            fig,
            _plots,
            _param_plots,
            _resp_plots,
            _colors,
            row_histograms,
        )

        return fig

    @app.callback(
        Output(parent.uuid("info-text"), "children"),
        [
            Input(parent.uuid("correlation-store-active-resp-param"), "data"),
        ],
    )
    def update_info_text(
        active_resp_param: Dict,
    ) -> List[Component]:
        if not (active_resp_param["response"] and active_resp_param["parameter"]):
            return []
        info_text = []
        active_response = active_resp_param["response"]

        active_index_text = "NA"
        active_index = active_response.get("index")
        if active_index is not None:
            active_index_text = _format_index_text(active_index)
        active_response_text = active_response.get("name")
        active_parameter_text = active_resp_param["parameter"]
        info_text.append(_generate_active_info_piece("RESPONSE", active_response_text))
        info_text.append(_generate_active_info_piece("INDEX", active_index_text))
        info_text.append(
            _generate_active_info_piece("PARAMETER", active_parameter_text)
        )
        return info_text

    @app.callback(
        Output(parent.uuid("correlation-store-active-resp-param"), "data"),
        [
            Input(
                parent.uuid("response-heatmap"),
                "clickData",
            ),
            Input(parent.uuid("parameter-selection-store-resp"), "data"),
            Input(parent.uuid("parameter-selection-store-param"), "data"),
            Input(parent.uuid("correlation-store-selected-obs"), "data"),
        ],
        [
            State(parent.uuid("correlation-store-active-resp-param"), "data"),
        ],
    )
    def update_active_resp_index_and_param(
        click_data: Dict,
        responses: List[str],
        parameters: List[str],
        selected_obs: Dict,
        active_resp_param: Dict,
    ) -> Dict:
        """
        update_active_resp_param is a callback function responsible for handling
        which response index and parameter are active.

        The correlation plots display correlation between a given response(at an index)
        and parameter for a chosen ensemble.

        It should be triggered upon
        - modifying the selection of parameters or responses
        - clicking on the heatmap, effectively choosing a combination of active
          response at an index and parameter
        - selection a new range in the response observation selector

        To avoid saying "response or parameter", we refer to either
        collectively by "variable".

        When the first variable is selected, they become active.
        Further selection of variables does not modify the active variable.
        Deselection of the active variable makes the first selected variable
        active, otherwise deselection does not affect the active state.

        Clicking on the heatmap makes the combination of response index and
        parameter active that are represented by the area of the heatmap that
        was clicked.
        """
        ctx = dash.callback_context
        triggered_id = ctx.triggered[0]["prop_id"].split(".")[0]
        if not triggered_id:
            raise PreventUpdate
        if triggered_id == parent.uuid("parameter-selection-store-resp"):
            if not responses:
                active_resp_param["response"] = None
            elif selected_obs and (
                not active_resp_param["response"]
                or not active_resp_param["response"]["name"] in responses
            ):
                response_name = next(iter(selected_obs.keys()))
                active_resp_param["response"] = {
                    "name": response_name,
                    "index": selected_obs[response_name][0],
                }
            elif selected_obs:
                response_name = active_resp_param["response"]["name"]
                response_index = active_resp_param["response"]["index"]
                if response_index not in selected_obs[response_name]:
                    active_resp_param["response"]["index"] = selected_obs[
                        response_name
                    ][0]

        elif triggered_id == parent.uuid("parameter-selection-store-param"):
            if not parameters:
                active_resp_param["parameter"] = None
            elif (
                not active_resp_param["parameter"]
                or not active_resp_param["parameter"] in parameters
            ):
                active_resp_param["parameter"] = parameters[0]
        elif triggered_id == parent.uuid("correlation-store-selected-obs"):
            if active_resp_param["response"] and selected_obs:
                response_name = active_resp_param["response"]["name"]
                if response_name in selected_obs:
                    active_resp_param["response"]["index"] = selected_obs[
                        response_name
                    ][0]
                else:
                    response_name = next(iter(selected_obs.keys()))
                    active_resp_param["response"] = {
                        "name": response_name,
                        "index": selected_obs[response_name][0],
                    }
        elif click_data:
            if "::" in click_data["points"][0]["x"]:
                resp_name, index = click_data["points"][0]["x"].split("::", maxsplit=1)
                index = _format_index_value(index)
            else:
                resp_name = click_data["points"][0]["x"]
                index = 0

            active_resp_param["parameter"] = click_data["points"][0]["y"]
            active_resp_param["response"] = {"name": resp_name, "index": index}
        elif not parameters and not responses:
            active_resp_param["parameter"] = None
            active_resp_param["response"] = None

        parent.save_state("active_correlation", active_resp_param)
        return active_resp_param


def _get_selected_ensembles_from_store(
    ensemble_selection_store: Dict[str, List],
) -> Optional[List[str]]:
    if ensemble_selection_store:
        return [
            selection["value"] for selection in ensemble_selection_store["selected"]
        ]
    return None


def _sort_dataframe(
    dataframe: pd.core.frame.DataFrame,
    index: Optional[pd.core.indexes.base.Index],
    sort_by_column: str,
) -> Tuple[pd.core.frame.DataFrame, pd.core.indexes.base.Index]:
    if index is None:
        index = dataframe[sort_by_column].abs().sort_values().index.copy()
    dataframe = dataframe.reindex(index)
    return dataframe, index


def _define_style_ensemble(index: int, x_axis: pd.Index) -> Dict:
    if str(x_axis[0]).isnumeric():
        style = deepcopy(assets.ERTSTYLE["response-plot"]["response-index"])
    else:
        style = deepcopy(assets.ERTSTYLE["response-plot"]["response"])
    ensemble_color = assets.get_color(index=index)
    style.update({"marker": {"color": ensemble_color}})
    style.update({"line": {"color": ensemble_color}})

    return style


def _layout_figure(x_axis_label: str) -> dict:
    layout = assets.ERTSTYLE["figure"]["layout"].copy()
    layout.update(dict(showlegend=False))
    layout.update(clickmode="event+select")
    layout.update({"xaxis": {"title": {"text": x_axis_label}}})
    layout.update(assets.ERTSTYLE["figure"]["layout-value-y-axis-label"])

    return layout


def _format_index_value(raw_index: Union[str, int]) -> Union[str, int]:
    try:
        return int(raw_index)
    except (TypeError, ValueError):
        return raw_index


def _format_index_text(raw_index: Union[str, int]) -> str:
    formatted_index = _format_index_value(raw_index)
    if isinstance(formatted_index, int):
        return str(formatted_index)
    else:
        try:
            return str(pd.Timestamp(raw_index).date())
        except (TypeError, ValueError):
            return str(raw_index)


def _create_scatterplot_specs(
    n_rows: int,
    histogram_row_span: int,
    spacer_row_span: int,
) -> List[List[Optional[Dict[str, int]]]]:
    specs: List[List[Optional[Dict[str, int]]]] = [
        [
            {
                "colspan": 2,
                "rowspan": n_rows - spacer_row_span - histogram_row_span,
            },
            None,
        ]
    ]
    # insert blanks into rows between scatterplot and histograms
    # sadly mypy cannot deal with this short hand notation
    specs += (n_rows - 1 - histogram_row_span) * [[None, None]]  # type:ignore
    specs.append([{"rowspan": histogram_row_span}, {"rowspan": histogram_row_span}])
    # insert blanks under histograms
    specs += (histogram_row_span - 1) * [[None, None]]  # type:ignore
    return specs


def _set_scatterplot_axes_titles(
    fig: go.Figure,
    row_histograms: int,
    active_response: str,
    active_parameter: str,
    x_value: Union[str, np.number],
) -> None:
    fig.update_xaxes(title_text=f"{active_response} @ {x_value}", row=1, col=1)
    fig.update_yaxes(title_text=f"{active_parameter}", row=1, col=1)
    fig.update_xaxes(title_text=f"{active_parameter}", row=row_histograms, col=1)
    fig.update_xaxes(title_text=f"{active_response}", row=row_histograms, col=2)


def _style_scatterplot(
    fig: go.Figure,
) -> None:
    fig.update_layout(assets.ERTSTYLE["figure"]["layout"])
    fig.update_layout(showlegend=False)


def _add_plots_to_multiplot(
    fig: go.Figure,
    _plots: List[PlotModel],
    _param_plots: Dict[str, np.ndarray],
    _resp_plots: Dict[str, np.ndarray],
    _colors: Dict[str, str],
    row_histograms: int,
) -> None:
    for plot in _plots:
        fig.add_trace(plot.repr, 1, 1)

    for ensemble_name in _param_plots:
        fig.add_trace(
            {
                "type": "histogram",
                "name": f"{ensemble_name}",
                "hoverlabel": {
                    "namelength": -1,
                },
                "x": _param_plots[ensemble_name],
                "showlegend": False,
                "marker_color": _colors[ensemble_name],
            },
            row_histograms,
            1,
        )
    for ensemble_name in _resp_plots:
        fig.add_trace(
            {
                "type": "histogram",
                "name": f"{ensemble_name}",
                "hoverlabel": {
                    "namelength": -1,
                },
                "x": _resp_plots[ensemble_name],
                "showlegend": False,
                "marker_color": _colors[ensemble_name],
            },
            row_histograms,
            2,
        )


def _generate_active_info_piece(title: str, value: str) -> Component:
    return html.Span(
        [
            f"{title}: ",
            dbc.Badge(
                value,
            ),
        ]
    )


def _get_selected_indexes(plots: List[PlotModel], data_range: Optional[Dict]) -> Dict:
    selected_indexes: Dict = {}
    if not data_range:
        return selected_indexes
    time_range = data_range.get("x")
    index_range = data_range.get("x2")
    for plot in plots:
        if plot.axis_type == AxisType.INDEX and index_range:
            indexes_in_range = plot.indexes_in_range(index_range)
        elif plot.axis_type == AxisType.TIMESTAMP and time_range:
            indexes_in_range = plot.indexes_in_range(time_range)
        else:
            continue
        if indexes_in_range:
            selected_indexes[plot.name] = indexes_in_range

    return selected_indexes


def _observation_index_plot(obs: Any, name: str, plot_idx: int) -> PlotModel:
    style = deepcopy(assets.ERTSTYLE["observation-selection-plot"])
    marker_color = assets.get_color(plot_idx)
    style.update(
        dict(
            marker={"color": marker_color, "size": 9},
            xaxis="x2" if obs.axis_type == AxisType.INDEX else "x",
        )
    )
    return PlotModel(
        x_axis=obs.axis,
        y_axis=[str(plot_idx) for _ in range(len(obs.axis))],
        text=obs.name,
        name=name,
        **style,
    )
